import warnings
import numpy as np
from scipy.spatial.distance import pdist, squareform
from .wmean import wmean


def knnimpute(x, k=3):
    """kNN missing value imputation using Euclidean distance.

    Parameters
    ----------
    x: array-like
        An array-like object that contains the data with NaNs.

    k: positive integer excluding 0, (default 3)
        The number of nearest neighbours to use.

    Returns
    -------
    z: array-like
        An array-like object corresponding to x with NaNs imputed.
    """
    
    # Tranpose x so we treat columns as features, and rows as samples
    x = x.T
    
    # Error check for k value
    if type(k) is not int:
        raise ValueError("k is not an integer")
    if k < 1:
        raise ValueError("k must be greater than zero")
    k_max = x.shape[1] - 1
    if k_max < k:
        raise ValueError("k value is too high. Max k value is {}".format(k_max))

    # z is the returned array with NaNs imputed
    z = x.copy()

    # Use columns without NaNs for knnimpute
    nan_check = np.isnan(x)
    no_nan = np.where(sum(nan_check.T) == 0, 1, 0) 

    # Error check that not all columns have NaNs
    x_no_nan = x[no_nan == 1]
    if x_no_nan.size == 0:
        raise ValueError("All colummns of the input data contain missing values. Unable to impute missing values.")

    # Calculate pairwise distances between columns, and covert to square-form distance matrix
    pair_dist = pdist(x_no_nan.T, metric="euclidean")
    sq_dist = squareform(pair_dist)

    # Make diagonals negative and sort
    dist = np.sort(sq_dist - np.eye(sq_dist.shape[0], sq_dist.shape[1])).T
    dist_idx = np.argsort(sq_dist - np.eye(sq_dist.shape[0], sq_dist.shape[1])).T

    # Find where neighbours are equal distance
    equal_dist_a = np.diff(dist[1:].T, 1, 1).T == 0
    equal_dist_a = equal_dist_a.astype(int)  # Convert to integer
    equal_dist_b = np.zeros(len(dist))
    equal_dist = np.concatenate((equal_dist_a, [equal_dist_b]))  # Concatenate

    # Get rows and cols for missing values
    nan_idx = np.argwhere(nan_check)
    nan_rows = nan_idx[:, 0]
    nan_cols = nan_idx[:, 1]
    # Make sure rows/cols are in a list (note: this happens when there is 1 missing value)
    if type(nan_rows) is not np.ndarray:
        nan_rows = [nan_rows]
        nan_cols = [nan_cols]

    # Impute each NaN value
    for i in range(len(nan_rows)):

        # Error check for rows with all NaNs
        if np.isnan(x[nan_rows[i], :]).all() == True:
            warnings.warn("Row {} contains all NaNs, so Row {} is imputed with zeros.".format(nan_rows[i], nan_rows[i]), Warning)

        # Create a loop from 1 to len(dist_idx) - k
        lastk = len(dist_idx) - k
        loopk = [1]
        while lastk > loopk[-1]:
            loopk.append(loopk[-1] + 1)

        # Impute
        for j in loopk:
            L_a = equal_dist[j + k - 2 :, nan_cols[i]]
            L = np.where(L_a == 0)[0][0]  # equal_dist neighbours

            x_vals_r = nan_rows[i]
            x_vals_c = dist_idx[j : j + k + L, nan_cols[i]]
            x_vals = x[x_vals_r, x_vals_c]
            weights = 1 / dist[1 : k + L + 1, nan_cols[i]]
            imp_val = wmean(x_vals, weights)  # imputed value
            if imp_val is not np.nan:
                z[nan_rows[i], nan_cols[i]] = imp_val
                break
                
    # Transpose z 
    z = z.T
    return z



import numpy as np
import pandas as pd
import scipy.stats
from statsmodels.stats.multitest import multipletests
from .table_check import table_check


def univariate_2class(DataTable, PeakTable, group, posclass, parametric=True, seed=None):
    """Creates a table of univariate statistics (2 class).

    Parameters
    ----------
    DataTable: DataFrame
        Data sheet with the required columns.

    PeakTable: DataFrame
        Peak sheet with the required columns.

    group: string
        Name of the column in the DataTable that contains the Class data.

    posclass: number or string
        Name of the positive class in the group column.

    parametric: boolean (default True)
        If parametric is False, Median 95%CI and Mann-Whitney U test is calculated instead of Mean 95%CI and TTest.

    seed: integer or None (default None)
        Used to seed the generator for the Median 95% CI bootstrap (resample with replacement). Note, if parametric=True the seed is irrelevant as Mean 95% CI is calculated parametrically.

    Returns
    -------
    StatsTable: DataFrame
        Table that contains multiple univariate statistics (2 class).
    """

    # Error checks
    table_check(DataTable, PeakTable, print_statement=False)
    if group not in DataTable:
        raise ValueError("Column '{}' does not exist in DataTable".format(group))
    if posclass not in DataTable[group].unique():
        raise ValueError("Positive class was not found in '{}' column.".format(group))
    if len(DataTable[group].unique()) != 2:
        raise ValueError("Column '{}' should have exactly 2 groups".format(group))

    # Get x0, x1
    peaklist = PeakTable["Name"]
    x = DataTable[peaklist]
    y = DataTable[group]
    x1 = x[y == posclass]  # x1 refers to posclass
    x0 = x[y != posclass]

    # Create stats table (include Idx, Name and Label)
    StatsTable = pd.DataFrame()
    StatsTable["Idx"] = PeakTable["Idx"]
    StatsTable["Name"] = PeakTable["Name"]
    StatsTable["Label"] = PeakTable["Label"]

    if parametric is True:
        # Calculate mean and std
        StatsTable["Grp0_Mean"] = np.nanmean(x0, axis=0)
        Mean095CI = (1.96 * (np.nanstd(x0, ddof=1, axis=0) / np.sqrt(x0.notnull().sum()))).values
        StatsTable["Grp0_Mean-95CI"] = list(zip(np.round(StatsTable["Grp0_Mean"] - Mean095CI, 2), np.round(StatsTable["Grp0_Mean"] + Mean095CI, 2)))
        StatsTable["Grp1_Mean"] = np.nanmean(x1, axis=0)
        Mean195CI = (1.96 * (np.nanstd(x1, ddof=1, axis=0) / np.sqrt(x1.notnull().sum()))).values
        StatsTable["Grp1_Mean-95CI"] = list(zip(np.round(StatsTable["Grp1_Mean"] - Mean195CI, 2), np.round(StatsTable["Grp1_Mean"] + Mean195CI, 2)))

        # Sign
        sign = np.nanmedian(x1, axis=0) / np.nanmedian(x0, axis=0)
        sign = np.where(sign > 1, 1, 0)
        StatsTable["Sign"] = sign

        # T test
        xnp = x.values
        t = scipy.stats.ttest_ind(x0, x1, nan_policy="omit")
        StatsTable["TTestStat"] = t[0]
        StatsTable["TTestPvalue"] = t[1]

        # BH correction
        bonTTest = multipletests(t[1], method="fdr_bh")
        # Round BH correction p-value
        StatsTable["bhQvalue"] = bonTTest[1]
    else:
        # Calculate median
        StatsTable["Grp0_Median"] = np.nanmedian(x0, axis=0)
        np.random.seed(seed)  # set seed for bootstrap
        x0_boot = []
        for i in range(100):
            x0_copy = x0.copy()
            x0_copy = x0_copy.iloc[np.random.randint(0, len(x0_copy), size=len(x0_copy))]
            x0_boot.append(np.nanmedian(x0_copy, axis=0))
        x0low = np.round(np.percentile(x0_boot, 2.5, axis=0), 2)
        x0high = np.round(np.percentile(x0_boot, 97.5, axis=0), 2)
        StatsTable["Grp0_Median-95CI"] = list(zip(x0low, x0high))

        StatsTable["Grp1_Median"] = np.nanmedian(x1, axis=0)
        np.random.seed(seed)  # set seed for bootstrap
        x1_boot = []
        for i in range(100):
            x1_copy = x1.copy()
            x1_copy = x1_copy.iloc[np.random.randint(0, len(x1_copy), size=len(x1_copy))]
            x1_boot.append(np.nanmedian(x1_copy, axis=0))
        x1low = np.round(np.percentile(x1_boot, 2.5, axis=0), 2)
        x1high = np.round(np.percentile(x1_boot, 97.5, axis=0), 2)
        StatsTable["Grp1_Median-95CI"] = list(zip(x1low, x1high))
        StatsTable["MedianFC"] = np.nanmedian(x1, axis=0) / np.nanmedian(x0, axis=0)

        # Sign
        sign = np.nanmedian(x1, axis=0) / np.nanmedian(x0, axis=0)
        sign = np.where(sign > 1, 1, 0)
        StatsTable["Sign"] = sign

        # Man-Whitney U
        m = []
        for i in range(len(x.T)):
            manwstat = scipy.stats.mannwhitneyu(x0.values[:, i], x1.values[:, i])
            m.append(manwstat)
        StatsTable["MannWhitneyU"] = [i[0] for i in m]
        StatsTable["MannWhitneyPvalue"] = [i[1] for i in m]

        # BH correction
        bonMannWhitney = multipletests(StatsTable["MannWhitneyPvalue"], method="fdr_bh")
        StatsTable["bhQvalue"] = bonMannWhitney[1]

    # Calculate total missing and total missing %
    nannum = x.isnull().sum()
    nanperc = nannum / x.shape[0]
    StatsTable["TotalMissing"] = nannum.values
    StatsTable["PercTotalMissing"] = np.round(nanperc.values * 100, 3)

    # Calculating missing % for group 0, and group 1...
    nanperc_0 = x0.isnull().sum() / len(x0)
    nanperc_1 = x1.isnull().sum() / len(x1)
    StatsTable["Grp0_Missing"] = np.round(nanperc_0.values * 100, 3)
    StatsTable["Grp1_Missing"] = np.round(nanperc_1.values * 100, 3)

    # Shapiro-Wilk
    s = []
    for i in range(len(x.T)):
        newx = x.values[:, i][~np.isnan(x.values[:, i])]
        shapstat = scipy.stats.shapiro(newx)
        s.append(shapstat)
    StatsTable["ShapiroW"] = [i[0] for i in s]  # Shapstat is a list of tuples, this is an approach to get this out of the tuple
    StatsTable["ShapiroPvalue"] = [i[1] for i in s]

    # Levenes
    l = []
    for i in range(len(x.T)):
        newx0 = x0.values[:, i][~np.isnan(x0.values[:, i])]
        newx1 = x1.values[:, i][~np.isnan(x1.values[:, i])]
        levstat = scipy.stats.levene(newx0, newx1)
        l.append(levstat)
    StatsTable["LeveneW"] = [i[0] for i in l]
    StatsTable["LevenePvalue"] = [i[1] for i in l]

    # Make the Idx column start from 1
    StatsTable.index = np.arange(1, len(StatsTable) + 1)
    return StatsTable



import numpy as np
import pandas as pd
from sklearn import decomposition
from bokeh.plotting import output_notebook, show
from bokeh.layouts import gridplot
from .scatter import scatter
from ..utils import ci95_ellipse


def pca(X, pcx=1, pcy=2, group_label=None, sample_label=None, peak_label=None):
    """Creates a PCA scores and loadings plot using Bokeh.

    Required Parameters
    -------------------
    X : array-like, shape = [n_samples]
        Inpute data
    """

    # Set model
    model = decomposition.PCA()
    model.fit(X)
    scores_ = model.transform(X)
    explained_var_ = model.explained_variance_ratio_ * 100

    # Extract scores, explained variance, and loadings for pcx and pcy
    x_score = scores_[:, (pcx - 1)]
    y_score = scores_[:, (pcy - 1)]
    x_expvar = explained_var_[(pcx - 1)]
    y_expvar = explained_var_[(pcy - 1)]
    x_load = model.components_[(pcx - 1), :]
    y_load = model.components_[(pcy - 1), :]

    # Colour for fig_score
    if group_label is None:
        col = ["blue", "green", "red"]
    else:
        col = None

    # Ensure group_label is an np.array
    group_label = np.array(group_label)

    # Scores plot
    fig_score = scatter(x_score, y_score, group=group_label, label=sample_label, size=5, xlabel="PC {} ({:0.1f}%)".format(pcx, x_expvar), ylabel="PC {} ({:0.1f}%)".format(pcy, y_expvar), title="PCA Score Plot (PC{} vs. PC{})".format(pcx, pcy), font_size="15pt", width=490, height=430, hover_xy=False, col_palette=col)
    
    # Loadings plot
    fig_load = scatter(x_load, y_load, size=7, label=peak_label, xlabel="PC {} ({:0.1f}%)".format(pcx, x_expvar), ylabel="PC {} ({:0.1f}%)".format(pcy, y_expvar), title="PCA Loadings Plot (PC{} vs. PC{})".format(pcx, pcy), font_size="15pt", width=490, height=430, hover_xy=False, shape="triangle", legend=False, hline=True, vline=True)

    # if group is None
    if group_label is None:
        group_label = [0] * len(X)

    # Score plot extra: 95% confidence ellipse using PCA
    unique_group = np.sort(np.unique(group_label))

    # Set colour per group
    list_color = ["red", "blue", "green", "black"]
    while len(list_color) < len(unique_group):  # Loop over list_color if number of groups > len(list_colour)
        list_color += list_color

    # Add 95% confidence ellipse for each unique group in a loop
    for i in range(len(unique_group)):
        # Get scores for the corresponding group
        group_i_x = []
        group_i_y = []
        for j in range(len(group_label)):
            if group_label[j] == unique_group[i]:
                group_i_x.append(x_score[j])
                group_i_y.append(y_score[j])

        # Calculate ci95 ellipse for each group
        data_circ_group = pd.DataFrame({"0": group_i_x, "1": group_i_y})
        m, outside_m = ci95_ellipse(data_circ_group, type="mean")
        p, outside_p = ci95_ellipse(data_circ_group, type="pop")

        # Plot ci95 ellipse outer line
        fig_score.line(m[:, 0], m[:, 1], color=list_color[i], line_width=2, alpha=0.8, line_dash="solid")
        fig_score.line(p[:, 0], p[:, 1], color=list_color[i], alpha=0.4)

        # Plot ci95 ellipse shade
        fig_score.patch(m[:, 0], m[:, 1], color=list_color[i], alpha=0.07)
        fig_score.patch(p[:, 0], p[:, 1], color=list_color[i], alpha=0.01)

    # Output this figure with fig_score and fig_load
    output_notebook()
    fig = gridplot([[fig_score, fig_load]])
    show(fig)



import numpy as np


def table_check(DataTable, PeakTable, print_statement=True):
    """Error checking for DataTable and PeakTable (used in load_dataXL).

    Parameters
    ----------
    DataTable: DataFrame
        Data sheet with the required columns.

    PeakTable: DataFrame
        Peak sheet with the required columns.

    print_statement: boolean (default True)
        If the error checks are successful and print_statement is True, the following is printed: "Data Table & Peak Table is suitable."
    """

    # Check DataTable for Idx, Class and SampleID
    data_columns = DataTable.columns.values

    if "Idx" not in data_columns:
        raise ValueError("Data Table does not contain the required 'Idx' column")
    if DataTable.Idx.isnull().values.any() == True:
        raise ValueError("Data Table Idx column cannot contain missing values")
    if len(np.unique(DataTable.Idx)) != len(DataTable.Idx):
        raise ValueError("Data Table Idx numbers are not unique. Please change")

    if "Class" not in data_columns:
        raise ValueError("Data Table does not contain the required 'Class' column")

    if "SampleID" not in data_columns:
        raise ValueError("Data Table does not contain the required 'SampleID' column")

    # Check PeakTable for Idx, Name, Label
    peak_columns = PeakTable.columns.values

    if "Idx" not in peak_columns:
        raise ValueError("Peak Table does not contain the required 'Idx' column")
    if PeakTable.Idx.isnull().values.any() == True:
        raise ValueError("Peak Table Idx column cannot contain missing values")
    if len(np.unique(PeakTable.Idx)) != len(PeakTable.Idx):
        raise ValueError("Peak Table Idx numbers are not unique. Please change")

    if "Name" not in peak_columns:
        raise ValueError("Peak Table does not contain the required 'Name' column")
    if PeakTable.Idx.isnull().values.any() == True:
        raise ValueError("Peak Table Name column cannot contain missing values")
    if len(np.unique(PeakTable.Idx)) != len(PeakTable.Idx):
        raise ValueError("Peak Table Name numbers are not unique. Please change")

    if "Label" not in peak_columns:
        raise ValueError("Data Table does not contain the required 'Label' column")

    # Check that Peak Names in PeakTable & DataTable match
    peak_list = PeakTable.Name
    data_columns = DataTable.columns.values
    temp = np.intersect1d(data_columns, peak_list)

    if len(temp) != len(peak_list):
        raise ValueError("The Peak Names in Data Table should exactly match the Peak Names in Peak Table. Remember that all Peak Names should be unique.")

    if print_statement is True:
        print("Data Table & Peak Table is suitable.")



import numpy as np


def scale(x, axis=0, ddof=1, method="auto", mu="default", sigma="default", return_mu_sigma=False):
    """Scales x (which can include nans) with method: 'auto', 'pareto', 'vast', or 'level'.

    Parameters
    ----------
    x: array-like
        An array-like object that contains the data.

    axis: integer or None, (default 0)
        The axis along which to operate

    ddof: integer, (default 1)
        The degrees of freedom correction. Note, by default ddof=1 unlike scipy.stats.zscore with ddof=0.

    method: string, (default "auto")
        Method used to scale x. Accepted methods are 'auto', 'pareto', 'vast' and 'level'.

    mu: number or "default", (default "default")
        If mu is provided it is used, however, by default it is calculated.

    sigma: number or "default",  (default "default")
        If sigma is provided it is used, however, by default it is calculated.

    return_mu_sigma: boolean, (default False)
        If return_mu_sigma is True, mu and sigma are returned instead of z. Note, this is useful if mu and sigma want to be stored for future use.

    Returns if return_mu_sigma = False
    ----------------------------------
    z: array-like
        An array-like object that contains the scaled data.

    Returns if return_mu_sigma = True
    ---------------------------------
    mu: number
        Calculated mu for x given axis and ddof.

    sigma: number
        Calculated sigma for x given axis and ddof.
    """

    x = np.array(x)

    # Simplier if we tranpose X if axis=1 (return x.T after the calculations)
    if axis == 1:
        x = x.T

    # Expand dimension if array is 1d
    if x.ndim == 1:
        x = np.expand_dims(x, axis=1)

    # Calculate mu and sigma if set to 'default' (ignoring nans)
    if mu is "default":
        mu = np.nanmean(x, axis=0)
    if sigma is "default":
        sigma = np.nanstd(x, axis=0, ddof=ddof)
        sigma = np.where(sigma == 0, 1, sigma)  # if a value in sigma equals 0 it is converted to 1

    # Error check before scaling
    if len(mu) != len(x.T):
        raise ValueError("Length of mu array does not match x matrix.")
    if len(sigma) != len(x.T):
        raise ValueError("Length of sigma array does not match x matrix.")

    # Scale based on selected method
    if method is "auto":
        z = (x - mu) / sigma
    elif method is "pareto":
        z = (x - mu) / np.sqrt(sigma)
    elif method is "vast":
        z = ((x - mu) / sigma) * (mu / sigma)
    elif method is "level":
        z = (x - mu) / mu
    else:
        raise ValueError("Method has to be either 'auto', 'pareto', 'vast', or 'level'.")

    # Return x.T if axis = 1
    if axis == 1:
        z = z.T

    if return_mu_sigma is True:
        return mu, sigma
    else:
        return z



import math
import numpy as np
from sklearn.decomposition import PCA


def ci95_ellipse(data, type="pop"):
    """ Construct a 95% confidence ellipse using PCA.

    Parameters
    ----------
    data : array-like, shape = [n_samples, 2]
        data[:,0] must represent x coordinates
        data[:,1] must represent y coordinates

    type : string, optional (default='pop')
        It must be 'pop' or 'mean'

    Returns
    -------
    ellipse: array-like, shape = [100, 2]
        ellipse[:,0] represents x coordinates of ellipse
        ellipse[:,1] represents y coordinates of ellipse

    outside: array-like, shape = [n_samples, 1]
        returns an 1d array (of 0/1) with length n_samples
        0 : ith sample is outside of ellipse
        1 : ith sample is inside of ellipse
    """

    # Build and fit PCA model
    pca = PCA()
    pca.fit(data)
    coeff = pca.components_
    score = pca.transform(data)
    eigvals = pca.explained_variance_

    # Calculate rotation angle
    phi = math.atan2(coeff[0, 1], coeff[0, 0])

    # This angle is between -pi and pi.
    # Let's shift it such that the angle is between 0 and 2pi
    if phi < 0:
        phi += 2 * math.pi

    # Get the coordinates of the data mean
    n = len(data)
    m = np.mean(data, axis=0)
    x0 = m[0]
    y0 = m[1]

    # Get the 95% confidence interval error ellipse
    # inverse of the chi-square cumulative distribution for  p = 0.05 & 2 d.f. = 5.9915
    chisquare_val = 5.9915
    if type is "pop":
        a = math.sqrt(chisquare_val * eigvals[0])
        b = math.sqrt(chisquare_val * eigvals[1])
    elif type is "mean":
        a = math.sqrt(chisquare_val * eigvals[0] / n)
        b = math.sqrt(chisquare_val * eigvals[1] / n)
    else:
        raise ValueError("type has to be 'pop' or 'mean'.")

    # the ellipse in x and y coordinates
    theta_grid = np.linspace(0, 2 * math.pi, num=100)
    ellipse_x_r = a * np.cos(theta_grid)
    ellipse_y_r = b * np.sin(theta_grid)

    # Define a rotation matrix
    R = np.array([[np.cos(phi), np.sin(phi)], [-np.sin(phi), np.cos(phi)]])
    # let's rotate the ellipse to some angle phi
    r_ellipse = np.dot(np.vstack((ellipse_x_r, ellipse_y_r)).T, R)

    # Draw the error ellipse
    x = r_ellipse[:, 0] + x0
    y = r_ellipse[:, 1] + y0
    ellipse = np.stack((x, y), axis=1)

    outside = []
    for i in range(len(score)):
        metric = (score[i, 0] / a) ** 2 + (score[i, 1] / b) ** 2
        if metric > 1:
            outside.append(1)
        else:
            outside.append(0)

    return ellipse, outside



import numpy as np
from scipy import stats
from bokeh.plotting import figure


def distribution(X, group, kde=True, title="Density Plot", xlabel="x", ylabel="Pr(x)", font_size="20pt", label_font_size="13pt", width=500, height=400, color_hist="green", color_kde="mediumturquoise"):
    """Creates a distribution plot using Bokeh.

    Required Parameters
    -------------------
    X : array-like, shape = [n_samples]
        Inpute data

    group : array-like, shape = [n_samples]
        Group label for sample
    """

    # Split into groups
    group_unique = np.sort(np.unique(group))
    x1 = X[group == group_unique[0]]
    x2 = X[group == group_unique[1]]
    if len(group_unique) == 4:
        x3 = X[group == group_unique[2]]
        x4 = X[group == group_unique[3]]

    # Density curve
    x1_min, x1_max = x1.min(), x1.max()
    x1_padding = (x1_max - x1_min) * 0.5
    x1_grid = np.linspace(x1_min - x1_padding, x1_max + x1_padding, 500)
    x1_pdf = stats.gaussian_kde(x1, "scott")
    x1_pdf_grid = x1_pdf(x1_grid)
    x1_grid = np.insert(x1_grid, 0, x1_min - x1_padding)
    x1_grid = np.insert(x1_grid, 0, x1_max + x1_padding)
    x1_pdf_grid = np.insert(x1_pdf_grid, 0, 0)
    x1_pdf_grid = np.insert(x1_pdf_grid, 0, 0)

    # Density curve
    x2_min, x2_max = x2.min(), x2.max()
    x2_padding = (x2_max - x2_min) * 0.5
    x2_grid = np.linspace(x2_min - x2_padding, x2_max + x2_padding, 500)
    x2_pdf = stats.gaussian_kde(x2, "scott")
    x2_pdf_grid = x2_pdf(x2_grid)
    x2_grid = np.insert(x2_grid, 0, x2_min - x2_padding)
    x2_grid = np.insert(x2_grid, 0, x2_max + x2_padding)
    x2_pdf_grid = np.insert(x2_pdf_grid, 0, 0)
    x2_pdf_grid = np.insert(x2_pdf_grid, 0, 0)

    # Density curve x3 and x4
    if len(group_unique) == 4:
        # Density curve
        x3_min, x3_max = x3.min(), x3.max()
        x3_padding = (x3_max - x3_min) * 0.5
        x3_grid = np.linspace(x3_min - x3_padding, x3_max + x3_padding, 500)
        x3_pdf = stats.gaussian_kde(x3, "scott")
        x3_pdf_grid = x3_pdf(x3_grid)
        x3_grid = np.insert(x3_grid, 0, x3_min - x3_padding)
        x3_grid = np.insert(x3_grid, 0, x3_max + x3_padding)
        x3_pdf_grid = np.insert(x3_pdf_grid, 0, 0)
        x3_pdf_grid = np.insert(x3_pdf_grid, 0, 0)

        # Density curve
        x4_min, x4_max = x4.min(), x4.max()
        x4_padding = (x4_max - x4_min) * 0.5
        x4_grid = np.linspace(x4_min - x4_padding, x4_max + x4_padding, 500)
        x4_pdf = stats.gaussian_kde(x4, "scott")
        x4_pdf_grid = x4_pdf(x4_grid)
        x4_grid = np.insert(x4_grid, 0, x4_min - x4_padding)
        x4_grid = np.insert(x4_grid, 0, x4_max + x4_padding)
        x4_pdf_grid = np.insert(x4_pdf_grid, 0, 0)
        x4_pdf_grid = np.insert(x4_pdf_grid, 0, 0)

    max_val_a = max(abs(max(x1_grid)), abs(min(x1_grid)))
    max_val_b = max(abs(max(x2_grid)), abs(min(x2_grid)))
    max_val_final = 0.2 * max(max_val_a, max_val_b)
    x_range_min = min(min(x1_grid) - max_val_final, min(x2_grid) - max_val_final)
    x_range_max = max(max(x1_grid) + max_val_final, max(x2_grid) + max_val_final)
    new_x_range = (x_range_min, x_range_max)
    new_y_range = (0, max(max(x1_pdf_grid) * 1.1, max(x2_pdf_grid) * 1.1))
    if len(group_unique) == 4:
        max_val_a = max(abs(max(x1_grid)), abs(min(x1_grid)))
        max_val_b = max(abs(max(x2_grid)), abs(min(x2_grid)))
        max_val_c = max(abs(max(x3_grid)), abs(min(x3_grid)))
        max_val_d = max(abs(max(x4_grid)), abs(min(x4_grid)))
        max_val_final = 0.1 * max(max_val_a, max_val_b, max_val_c, max_val_d)
        x_range_min = min(min(x1_grid) - max_val_final, min(x2_grid) - max_val_final, min(x3_grid) - max_val_final, min(x4_grid) - max_val_final)
        x_range_max = max(max(x1_grid) + max_val_final, max(x2_grid) + max_val_final, max(x3_grid) + max_val_final, min(x4_grid) + max_val_final)
        new_x_range = (x_range_min, x_range_max)
        new_y_range = (0, max(max(x1_pdf_grid) * 1.1, max(x2_pdf_grid) * 1.1, max(x3_pdf_grid) * 1.1, max(x4_pdf_grid) * 1.1))

    # Figure
    fig = figure(title=title, x_axis_label=xlabel, y_axis_label=ylabel, plot_width=width, plot_height=height, x_range=new_x_range, y_range=new_y_range)
    if kde is True:
        if len(group_unique) == 4:
            fig.patch(x1_grid, x1_pdf_grid, alpha=0.1, color="red", line_color="grey", line_width=1)
            fig.patch(x2_grid, x2_pdf_grid, alpha=0.1, color="blue", line_color="grey", line_width=1)
            fig.patch(x3_grid, x3_pdf_grid, alpha=0.2, color="red", line_color="grey", line_width=1)
            fig.patch(x4_grid, x4_pdf_grid, alpha=0.2, color="blue", line_color="grey", line_width=1)
        else:
            fig.patch(x1_grid, x1_pdf_grid, alpha=0.3, color="red", line_color="grey", line_width=1)
            fig.patch(x2_grid, x2_pdf_grid, alpha=0.3, color="blue", line_color="grey", line_width=1)

    # Y-axis should always start at 0
    fig.y_range.start = 0

    # Font-sizes
    fig.title.text_font_size = font_size
    fig.xaxis.axis_label_text_font_size = label_font_size
    fig.yaxis.axis_label_text_font_size = label_font_size

    # Extra padding
    fig.min_border_left = 20
    fig.min_border_right = 20
    fig.min_border_top = 20
    fig.min_border_bottom = 20

    return fig



import numpy as np


def wmean(x, weights):
    """Returns Weighted Mean. Ignores NaNs and handles infinite weights.

    Parameters
    ----------
    x: array-like [n_samples]
        An array-like object that contains the data.

    weights: array-like [n_samples]
        An array-like object that contains the corresponding weights.

    Returns
    ----------------------------------
    m: number
        The weighted mean.
    """

    # Flatten x and weights
    x = x.flatten()
    weights = weights.flatten()

    # Find NaNs
    nans = np.isnan(x)
    infs = np.isinf(weights)

    # If all x are nans, return np.nan
    if nans.all() == True:
        m = np.nan
        return m

    # If there are infinite weights, use the corresponding x
    if infs.any() == True:
        m = np.nanmean(x[infs])
        return m

    # Set NaNs to zero
    x[nans] = 0
    weights[nans] = 0
    
    # Normalize the weights + calculate Weighted Mean
    weights = weights / np.sum(weights)
    m = np.matmul(weights, x)
    return m



import numpy as np
from sklearn.metrics import confusion_matrix, roc_auc_score


def binary_metrics(y_true, y_pred, cut_off=0.5):
    """ Return a dict of binary stats with the following metrics: R2, auc, accuracy, precision, sensitivity, specificity, and F1 score.

    Parameters
    ----------
    y_true : array-like, shape = [n_samples]
        Binary label for samples (0s and 1s).

    y_pred : array-like, shape = [n_samples]
        Predicted y score for samples.

    cut_off : number, (default 0.5)
        A value for y_pred greater-than or equal to the cut_off will be treated as 1, otherwise it will be treated as 0 for the confusion matrix.

    Returns
    -------
    stats: dict
        dict containing calculated R2, auc, accuracy, precision, sensitivity, specificity, and F1 score.
    """

    # Convert to array
    y_true_arr = np.array(y_true)
    y_pred_arr = np.array(y_pred)

    # Error checks
    if y_true_arr.ndim != 1:
        raise ValueError("y_true should only have 1 dimension.")
    if y_pred_arr.ndim != 1:
        raise ValueError("y_pred should only have 1 dimension.")
    if y_true_arr.shape[0] != y_pred_arr.shape[0]:
        raise ValueError("The number of values in y_true should match y_pred.")
    if np.array_equal(sorted(set(y_true_arr)), [0, 1]) is False:
        raise ValueError("y_true should only contain 0s and 1s")

    # Get confusion matrix
    y_pred_round = np.where(y_pred_arr >= cut_off, 1, 0)
    tn, fp, fn, tp = confusion_matrix(y_true_arr, y_pred_round).ravel()

    # Binary statistics dictionary
    stats = {}
    stats["R²"] = 1 - (sum((y_true_arr - y_pred_arr) ** 2) / sum((y_true_arr - np.mean(y_true_arr)) ** 2))
    stats["AUC"] = roc_auc_score(y_true_arr, y_pred_arr)
    stats["ACCURACY"] = safe_div((tp + tn), (tp + tn + fp + fn))
    stats["PRECISION"] = safe_div((tp), (tp + fp))
    stats["SENSITIVITY"] = safe_div((tp), (tp + fn))
    stats["SPECIFICITY"] = safe_div((tn), (tn + fp))
    stats["F1-SCORE"] = safe_div((2 * tp), (2 * tp + fp + fn))
    return stats


def safe_div(a, b):
    """Return np.nan if the demoninator is 0."""
    if b == 0:
        return np.nan  
    return a / b



import numpy as np
import pandas as pd
from scipy import stats
from bokeh.plotting import ColumnDataSource, figure
from bokeh.models import HoverTool


def boxplot(X, group, violin=False, title="", xlabel="Group", ylabel="Value", font_size="20pt", label_font_size="13pt", width=500, height=400, color="whitesmoke", color_violin="mediumturquoise", width_boxplot=1, width_violin=1, y_range=None, group_name=None, group_name_sort=None):
    """Creates a boxplot using Bokeh.

    Required Parameters
    -------------------
    X : array-like, shape = [n_samples]
        Inpute data

    group : array-like, shape = [n_samples]
        Group label for sample
    """

    if group_name_sort is None:
        group_name_sort = [str(i) for i in list(set(group))]

    # unique groups as strings
    if group_name is None:
        group_name = [str(i) for i in list(set(group))]

    # Colors
    if isinstance(color, str):
        color = [color] * len(group_name)
    if isinstance(color_violin, str):
        color_violin = [color_violin] * len(group_name)

    # Combine X and group into pandas table
    X_pd = pd.Series(X, name="X")
    group_pd = pd.Series(group, name="group")
    table = pd.concat([X_pd, group_pd], axis=1)

    # Find the quartiles and IQR for each group
    X_groupby = table.groupby("group", sort=True)
    q1 = X_groupby.quantile(q=0.25)
    q2 = X_groupby.quantile(q=0.50)
    q3 = X_groupby.quantile(q=0.75)
    iqr = q3 - q1
    upper = q3 + 1.5 * iqr
    lower = q1 - 1.5 * iqr

    # Find the outliers for each category
    def outliers(group):
        group_name = group.name
        return group[(group.X > upper.loc[group_name]["X"]) | (group.X < lower.loc[group_name]["X"])]["X"]

    out = X_groupby.apply(outliers).dropna()

    # prepare outlier data for plotting, we need coordinates for every outlier.
    outx = []
    outy = []
    outidx = []
    if not out.empty:
        outx = []
        outy = []
        outidx = []
        for keys in out.index:
            outx.append(str(keys[0]))
            outy.append(out.loc[keys[0]].loc[keys[1]])
            outidx.append(keys[1])

    # If no outliers, shrink lengths of stems to be no longer than the minimums or maximums
    qmin = X_groupby.quantile(q=0.00)
    qmax = X_groupby.quantile(q=1.00)
    upper.X = [min([x, y]) for (x, y) in zip(list(qmax.loc[:, "X"]), upper.X)]
    lower.X = [max([x, y]) for (x, y) in zip(list(qmin.loc[:, "X"]), lower.X)]

    # Bokeh data source boxplot
    source_boxplot = ColumnDataSource(data=dict(group_name=group_name_sort, upper=upper.X, lower=lower.X, q1=q1.X, q2=q2.X, q3=q3.X, color=color))

    # Bokeh data source outlier
    source_outliers = ColumnDataSource(data=dict(outx=outx, outy=outy, outidx=outidx))

    # Set default y_range if None
    if y_range is None:
        max_val = max(abs(np.min(X)), abs(np.max(X)))
        violin_add = (np.max(X) - np.min(X)) * 0.15
        max_val_1 = 0.1 * max_val + violin_add
        y_range = (np.min(X) - max_val_1, np.max(X) + max_val_1)

    # Base figure
    fig = figure(title=title, x_range=group_name, x_axis_label=xlabel, y_axis_label=ylabel, plot_width=width, plot_height=height, y_range=y_range)

    # Plot probability density shade if violin is True
    if violin is True:
        vwidth = 0.5
        table["group"] = table.group.astype("category")
        for i in table["group"].cat.categories:
            y_data = table[table["group"] == i]["X"]
            y_min, y_max = y_data.min(), y_data.max()
            y_padding = (y_max - y_min) * 0.15
            y_grid = np.linspace(y_min - y_padding, y_max + y_padding, 60)
            pdf = stats.gaussian_kde(y_data, "scott")
            x_pdf = pdf(y_grid)
            x_pdf = x_pdf / x_pdf.max() * vwidth / (2 / width_violin)
            x_patch = np.append(x_pdf, -x_pdf[::-1])
            y_patch = np.append(y_grid, y_grid[::-1])
            for j in range(len(group_name)):
                if group_name[j] == str(i):
                    val = j
            fig.patch((x_patch + val + 0.5), y_patch, alpha=0.3, color=color_violin[val], line_color="grey", line_width=1)

    ## Boxplot (Stems, Boxes, Whiskers, and Outliers) ##
    # Stems
    stem1 = fig.segment("group_name", "upper", "group_name", "q3", line_color="black", source=source_boxplot)
    stem2 = fig.segment("group_name", "lower", "group_name", "q1", line_color="black", source=source_boxplot)

    # Boxes
    box1 = fig.vbar("group_name", width_boxplot / 10, "q2", "q3", fill_color="color", line_color="black", alpha=0.8, source=source_boxplot)
    box2 = fig.vbar("group_name", width_boxplot / 10, "q1", "q2", fill_color="color", line_color="black", alpha=0.8, source=source_boxplot)

    # Whiskers (almost-0 height rects)
    whisker1 = fig.rect("group_name", "lower", 0.05, 0.01, line_color="black", source=source_boxplot)
    whisker2 = fig.rect("group_name", "upper", 0.05, 0.01, line_color="black", source=source_boxplot)

    # Outliers
    outliers = fig.circle("outx", "outy", size=4, color="red", fill_alpha=0.4, source=source_outliers)

    # Hovertool for boxplot
    fig.add_tools(HoverTool(renderers=[stem1, stem2, box1, box2], tooltips=[("Upper", "@upper{1.11}"), ("Q3", "@q3{1.11}"), ("Median", "@q2{1.11}"), ("Q1", "@q1{1.11}"), ("Lower", "@lower{1.11}")]))

    # Hovertool for outliers
    fig.add_tools(HoverTool(renderers=[outliers], tooltips=[("Index", "@outidx"), (ylabel, "@outy")]))

    # Font-sizes
    fig.title.text_font_size = font_size
    fig.xaxis.axis_label_text_font_size = label_font_size
    fig.yaxis.axis_label_text_font_size = label_font_size

    # Extra padding
    fig.min_border_left = 20
    fig.min_border_right = 20
    fig.min_border_top = 20
    fig.min_border_bottom = 20

    return fig



import numpy as np
from bokeh.plotting import ColumnDataSource, figure
from bokeh.models import Slope, Span, HoverTool


def scatter(x, y, label=None, group=None, title="Scatter Plot", xlabel="x", ylabel="y", width=600, height=600, legend=True, size=4, shape="circle", font_size="16pt", label_font_size="13pt", col_palette=None, hover_xy=True, gradient=False, hline=False, vline=False, xrange=None, yrange=None):
    """Creates a scatterplot using Bokeh.

    Required Parameters
    -------------------
    x : array-like, shape = [n_samples]
        Inpute data for x-axis.

    y : array-like, shape = [n_samples]
        Inpute data for y-axis.
    """

    # Error check
    if len(x) != len(y):
        raise ValueError("length of X does not match length of Y.")

    # If label is None, give an index based on input order
    if label is None:
        label_copy = {}
        label_copy["Idx"] = list(range(len(x)))
    else:
        try:
            label2 = label.copy()
            label2_dict = label2.to_dict("series")
            label_copy = label2_dict  # Ensure I don't overwrite label (when plot_groupmean=True)
        except TypeError:
            label2 = label.copy()
            label_copy = {}
            label_copy[label2.name] = label2.values.tolist()

    # If colour palette is None (default):
    if col_palette is None:
        col_palette = ["red", "blue", "green"]

    # Group is None or allow for multiple classes (can add more in the Future)

    if group is None:
        group_copy = [None] * len(x)
        col = []
        for i in range(len(x)):
            col.append(col_palette[2])
    else:
        group_copy = group.copy()
        group_unique = np.sort(np.unique(group_copy))
        col = []
        for i in range(len(group_copy)):
            if group_copy[i] == group_unique[0]:
                col.append(col_palette[0])
            elif group_copy[i] == group_unique[1]:
                col.append(col_palette[1])
            else:
                col.append(col_palette[2])

    # Bokeh data source with data labels
    data = {"x": x, "y": y, "group": group_copy, "col": col}
    data_label = {}
    for name, val in label_copy.items():
        data_label[name] = val
    data.update(data_label)
    source = ColumnDataSource(data=data)

    # Tool-tip (add everything in label_copy)
    TOOLTIPS = []
    if hover_xy is True:
        TOOLTIPS = [("x", "@x{1.111}"), ("y", "@y{1.111}")]
    for name, val in data_label.items():
        TOOLTIPS.append((str(name), "@" + str(name)))

    # Base figure
    fig = figure(title=title, x_axis_label=xlabel, y_axis_label=ylabel, plot_width=width, plot_height=height, x_range=xrange, y_range=yrange)

    # Add to plot
    if shape is "circle":
        shape = fig.circle("x", "y", size=size, alpha=0.6, color="col", legend="group", source=source)
    elif shape is "triangle":
        shape = fig.triangle("x", "y", size=size, alpha=0.6, color="col", legend="group", source=source)
    else:
        raise ValueError("shape has to be either 'circle' or 'triangle'.")

    shape_hover = HoverTool(renderers=[shape], tooltips=TOOLTIPS)
    fig.add_tools(shape_hover)

    if gradient is not False:
        slope = Slope(gradient=gradient, y_intercept=0, line_color="black", line_width=2, line_alpha=0.3)
        fig.add_layout(slope)
        new_gradient = -(1 / gradient)
        slope2 = Slope(gradient=new_gradient, y_intercept=0, line_color="black", line_dash="dashed", line_width=2, line_alpha=0.10)
        fig.add_layout(slope2)

    if hline is not False:
        h = Span(location=0, dimension="width", line_color="black", line_width=3, line_alpha=0.15)
        fig.add_layout(h)

    if vline is not False:
        v = Span(location=0, dimension="height", line_color="black", line_width=3, line_alpha=0.15)
        fig.add_layout(v)

    # Font-sizes
    fig.title.text_font_size = font_size
    fig.xaxis.axis_label_text_font_size = label_font_size
    fig.yaxis.axis_label_text_font_size = label_font_size

    # Extra padding
    fig.min_border_left = 20
    fig.min_border_right = 20
    fig.min_border_top = 20
    fig.min_border_bottom = 20

    # Remove legend
    if legend is False:
        fig.legend.visible = False

    return fig



from abc import ABC, abstractmethod, abstractproperty
import numpy as np
import pandas as pd
import scipy
from bokeh.layouts import widgetbox, gridplot, column, row, layout
from bokeh.models import HoverTool, Band
from bokeh.models.widgets import DataTable, Div, TableColumn
from bokeh.models.annotations import Title
from bokeh.plotting import ColumnDataSource, figure, output_notebook, show
from scipy import interp
from sklearn import metrics
from sklearn.utils import resample
from ..bootstrap import Perc, BC, BCA
from ..plot import scatter, scatterCI, boxplot, distribution, permutation_test, roc_calculate, roc_plot
from ..utils import binary_metrics


class BaseModel(ABC):
    """Base class for models: PLS_SIMPLS."""

    @abstractmethod
    def __init__(self):
        pass

    @abstractmethod
    def train(self):
        """Trains the model."""
        pass

    @abstractmethod
    def test(self):
        """Tests the model."""
        pass

    @abstractproperty
    def bootlist(self):
        """A list of attributes for bootstrap resampling."""
        pass

    def evaluate(self, testset=None, specificity=False, cutoffscore=False, bootnum=1000):
        """Plots a figure containing a Violin plot, Distribution plot, ROC plot and Binary Metrics statistics.

        Parameters
        ----------
        testset : array-like, shape = [n_samples, 2] or None, (default None)
            If testset is None, use train Y and train Y predicted for evaluate. Alternatively, testset is used to evaluate model in the format [Ytest, Ypred].

        specificity : number or False, (default False)
            Use the specificity to draw error bar. When False, use the cutoff score of 0.5.

        cutoffscore : number or False, (default False)
            Use the cutoff score to draw error bar. When False, use the specificity selected.

        bootnum : a positive integer, (default 1000)
            The number of bootstrap samples used in the computation.
        """
        Ytrue_train = self.Y
        Yscore_train = self.Y_pred.flatten()

        # Get Ytrue_test, Yscore_test from testset
        if testset is not None:
            Ytrue_test = np.array(testset[0])
            Yscore_test = np.array(testset[1])

            # Error checking
            if len(Ytrue_test) != len(Yscore_test):
                raise ValueError("evaluate can't be used as length of Ytrue does not match length of Yscore in test set.")
            if len(np.unique(Ytrue_test)) != 2:
                raise ValueError("Ytrue_test needs to have 2 groups. There is {}".format(len(np.unique(Y))))
            if np.sort(np.unique(Ytrue_test))[0] != 0:
                raise ValueError("Ytrue_test should only contain 0s and 1s.")
            if np.sort(np.unique(Ytrue_test))[1] != 1:
                raise ValueError("Ytrue_test should only contain 0s and 1s.")

            # Get Yscore_combined and Ytrue_combined_name (Labeled Ytrue)
            Yscore_combined = np.concatenate([Yscore_train, Yscore_test])
            Ytrue_combined = np.concatenate([Ytrue_train, Ytrue_test + 2])  # Each Ytrue per group is unique
            Ytrue_combined_name = Ytrue_combined.astype(np.str)
            Ytrue_combined_name[Ytrue_combined == 0] = "Train (0)"
            Ytrue_combined_name[Ytrue_combined == 1] = "Train (1)"
            Ytrue_combined_name[Ytrue_combined == 2] = "Test (0)"
            Ytrue_combined_name[Ytrue_combined == 3] = "Test (1)"

        # Expliclity states which metric and value is used for the error_bar
        if specificity is not False:
            metric = "specificity"
            val = specificity
        elif cutoffscore is not False:
            metric = "cutoffscore"
            val = cutoffscore
        else:
            metric = "specificity"
            val = 0.8

        # ROC plot
        tpr, fpr, tpr_ci, stats, stats_bootci = roc_calculate(Ytrue_train, Yscore_train, bootnum=100, metric=metric, val=val)
        roc_title = "Specificity: {}".format(np.round(stats["val_specificity"], 2))
        roc_bokeh = roc_plot(tpr, fpr, tpr_ci, width=320, height=315, title=roc_title, errorbar=stats["val_specificity"])
        if testset is not None:
            fpr_test, tpr_test, threshold_test = metrics.roc_curve(Ytrue_test, Yscore_test, pos_label=1, drop_intermediate=False)
            fpr_test = np.insert(fpr_test, 0, 0)
            tpr_test = np.insert(tpr_test, 0, 0)
            roc_bokeh.line(fpr_test, tpr_test, color="red", line_width=3.5, alpha=0.6, legend="ROC Curve (Test)")  # Add ROC Curve(Test) to roc_bokeh

        # Violin plot
        violin_title = "Cut-off: {}".format(np.round(stats["val_cutoffscore"], 2))
        if testset is None:
            violin_bokeh = boxplot(Yscore_train, Ytrue_train, xlabel="Class", ylabel="Predicted Score", violin=True, color=["#FFCCCC", "#CCE5FF"], width=320, height=315, title=violin_title, font_size="11pt")
        else:
            violin_bokeh = boxplot(Yscore_combined, Ytrue_combined_name, xlabel="Class", ylabel="Predicted Score", violin=True, color=["#fcaeae", "#aed3f9", "#FFCCCC", "#CCE5FF"], width=320, height=315, group_name=["Train (0)", "Test (0)", "Train (1)", "Test (1)"], group_name_sort=["Test (0)", "Test (1)", "Train (0)", "Train (1)"], title=violin_title, font_size="11pt")
        violin_bokeh.multi_line([[-100, 100]], [[stats["val_cutoffscore"], stats["val_cutoffscore"]]], line_color="black", line_width=2, line_alpha=1.0, line_dash="dashed")

        # Distribution plot
        if testset is None:
            dist_bokeh = distribution(Yscore_train, group=Ytrue_train, kde=True, title="", xlabel="Predicted Score", ylabel="p.d.f.", width=320, height=315)
        else:
            dist_bokeh = distribution(Yscore_combined, group=Ytrue_combined_name, kde=True, title="", xlabel="Predicted Score", ylabel="p.d.f.", width=320, height=315)
        dist_bokeh.multi_line([[stats["val_cutoffscore"], stats["val_cutoffscore"]]], [[-100, 100]], line_color="black", line_width=2, line_alpha=1.0, line_dash="dashed")

        # Man-Whitney U for Table (round and use scienitic notation if p-value > 0.001)
        manw_pval = scipy.stats.mannwhitneyu(Yscore_train[Ytrue_train == 0], Yscore_train[Ytrue_train == 1], alternative="two-sided")[1]
        if manw_pval > 0.001:
            manw_pval_round = "%0.2f" % manw_pval
        else:
            manw_pval_round = "%0.2e" % manw_pval
        if testset is not None:
            testmanw_pval = scipy.stats.mannwhitneyu(Yscore_test[Ytrue_test == 0], Yscore_test[Ytrue_test == 1], alternative="two-sided")[1]
            if testmanw_pval > 0.001:
                testmanw_pval_round = "%0.2f" % testmanw_pval
            else:
                testmanw_pval_round = "%0.2e" % testmanw_pval

        # Create a stats table for test
        if testset is not None:
            teststats = binary_metrics(Ytrue_test, Yscore_test, cut_off=stats["val_cutoffscore"])
            teststats_round = {}
            for i in teststats.keys():
                teststats_round[i] = np.round(teststats[i], 2)

        # Round stats, and stats_bootci for Table
        stats_round = {}
        for i in stats.keys():
            stats_round[i] = np.round(stats[i], 2)
        bootci_round = {}
        for i in stats_bootci.keys():
            bootci_round[i] = np.round(stats_bootci[i], 2)

        # Create table
        tabledata = dict(
            evaluate=[["Train"]],
            manw_pval=[["{}".format(manw_pval_round)]],
            auc=[["{} ({}, {})".format(stats_round["AUC"], bootci_round["AUC"][0], bootci_round["AUC"][1])]],
            accuracy=[["{} ({}, {})".format(stats_round["ACCURACY"], bootci_round["ACCURACY"][0], bootci_round["ACCURACY"][1])]],
            precision=[["{} ({}, {})".format(stats_round["PRECISION"], bootci_round["PRECISION"][0], bootci_round["PRECISION"][1])]],
            sensitivity=[["{} ({}, {})".format(stats_round["SENSITIVITY"], bootci_round["SENSITIVITY"][0], bootci_round["SENSITIVITY"][1])]],
            specificity=[["{} ({}, {})".format(stats_round["SPECIFICITY"], bootci_round["SPECIFICITY"][0], bootci_round["SPECIFICITY"][1])]],
            F1score=[["{} ({}, {})".format(stats_round["F1-SCORE"], bootci_round["F1-SCORE"][0], bootci_round["F1-SCORE"][1])]],
            R2=[["{} ({}, {})".format(stats_round["R²"], bootci_round["R²"][0], bootci_round["R²"][1])]],
        )

        # Append test data
        if testset is not None:
            tabledata["evaluate"].append(["Test"])
            tabledata["manw_pval"].append([testmanw_pval_round])
            tabledata["auc"].append([teststats_round["AUC"]])
            tabledata["accuracy"].append([teststats_round["ACCURACY"]])
            tabledata["precision"].append([teststats_round["PRECISION"]])
            tabledata["sensitivity"].append([teststats_round["SENSITIVITY"]])
            tabledata["specificity"].append([teststats_round["SPECIFICITY"]])
            tabledata["F1score"].append([teststats_round["F1-SCORE"]])
            tabledata["R2"].append([teststats_round["R²"]])

        # Plot table
        source = ColumnDataSource(data=tabledata)
        columns = [TableColumn(field="evaluate", title="Evaluate"), TableColumn(field="manw_pval", title="MW-U Pvalue"), TableColumn(field="R2", title="R2"), TableColumn(field="auc", title="AUC"), TableColumn(field="accuracy", title="Accuracy"), TableColumn(field="precision", title="Precision"), TableColumn(field="sensitivity", title="Sensitivity"), TableColumn(field="F1score", title="F1score")]
        table_bokeh = widgetbox(DataTable(source=source, columns=columns, width=950, height=90), width=950, height=80)

        # Title
        if specificity is not False:
            title = "Specificity fixed to: {}".format(np.round(val, 2))
        elif cutoffscore is not False:
            title = "Score cut-off fixed to: {}".format(np.round(val, 2))
        else:
            title = "Specificity fixed to: {}".format(np.round(val, 2))
        title_bokeh = "<h3>{}</h3>".format(title)

        # Combine table, violin plot and roc plot into one figure
        fig = layout([[violin_bokeh, dist_bokeh, roc_bokeh], [table_bokeh]], toolbar_location="right")
        output_notebook()
        show(column(Div(text=title_bokeh, width=900, height=50), fig))

    def calc_bootci(self, bootnum=100, type="bca"):
        """Calculates bootstrap confidence intervals based on bootlist.

        Parameters
        ----------
        bootnum : a positive integer, (default 100)
            The number of bootstrap samples used in the computation.

        type : 'bc', 'bca', 'perc', (default 'bca')
            Methods for bootstrap confidence intervals. 'bc' is bias-corrected bootstrap confidence intervals. 'bca' is bias-corrected and accelerated bootstrap confidence intervals. 'perc' is percentile confidence intervals.
        """
        bootlist = self.bootlist
        if type is "bca":
            boot = BCA(self, self.X, self.Y, self.bootlist, bootnum=bootnum)
        if type is "bc":
            boot = BC(self, self.X, self.Y, self.bootlist, bootnum=bootnum)
        if type is "perc":
            boot = Perc(self, self.X, self.Y, self.bootlist, bootnum=bootnum)
        self.bootci = boot.run()

    def plot_featureimportance(self, PeakTable, peaklist=None, ylabel="Label", sort=True):
        """Plots feature importance metrics.

        Parameters
        ----------
        PeakTable : DataFrame
            Peak sheet with the required columns.

        peaklist : list or None, (default None)
            Peaks to include in plot (the default is to include all samples).

        ylabel : string, (default "Label")
            Name of column in PeakTable to use as the ylabel.

        sort : boolean, (default True)
            Whether to sort plots in absolute descending order.

        Returns
        -------
        Peaksheet : DataFrame
            New PeakTable with added "Coef" and "VIP" columns (+ "Coef-95CI" and  "VIP-95CI" if calc_bootci is used prior to plot_featureimportance).
        """
        if not hasattr(self, "bootci"):
            print("Use method calc_bootci prior to plot_featureimportance to add 95% confidence intervals to plots.")
            ci_coef = None
            ci_vip = None
        else:
            ci_coef = self.bootci["model.coef_"]
            ci_vip = self.bootci["model.vip_"]

        # Remove rows from PeakTable if not in peaklist
        if peaklist is not None:
            PeakTable = PeakTable[PeakTable["Name"].isin(peaklist)]
        peaklabel = PeakTable[ylabel]

        # Plot
        fig_1 = scatterCI(self.model.coef_, ci=ci_coef, label=peaklabel, hoverlabel=PeakTable[["Idx", "Name", "Label"]], hline=0, col_hline=True, title="Coefficient Plot", sort_abs=sort)
        fig_2 = scatterCI(self.model.vip_, ci=ci_vip, label=peaklabel, hoverlabel=PeakTable[["Idx", "Name", "Label"]], hline=1, col_hline=False, title="Variable Importance in Projection (VIP)", sort_abs=sort)
        fig = layout([[fig_1], [fig_2]])
        output_notebook()
        show(fig)

        # Return table with: Idx, Name, Label, Coefficient, 95CI, VIP, 95CI
        if not hasattr(self, "bootci"):
            coef = pd.DataFrame([self.model.coef_]).T
            coef.rename(columns={0: "Coef"}, inplace=True)
            vip = pd.DataFrame([self.model.vip_]).T
            vip.rename(columns={0: "VIP"}, inplace=True)
        else:
            coef = pd.DataFrame([self.model.coef_, self.bootci["model.coef_"]]).T
            coef.rename(columns={0: "Coef", 1: "Coef-95CI"}, inplace=True)
            vip = pd.DataFrame([self.model.vip_, self.bootci["model.vip_"]]).T
            vip.rename(columns={0: "VIP", 1: "VIP-95CI"}, inplace=True)

        Peaksheet = PeakTable.copy()
        Peaksheet["Coef"] = coef["Coef"].values
        Peaksheet["VIP"] = vip["VIP"].values
        if hasattr(self, "bootci"):
            Peaksheet["Coef-95CI"] = coef["Coef-95CI"].values
            Peaksheet["VIP-95CI"] = vip["VIP-95CI"].values
        return Peaksheet

    def permutation_test(self, nperm=100):
        """Plots permutation test figures.

        Parameters
        ----------
        nperm : positive integer, (default 100)
            Number of permutations.
        """
        fig = permutation_test(self, self.X, self.Y, nperm=nperm)
        output_notebook()
        show(fig)



import numpy as np
from abc import ABC, abstractmethod
from sklearn.model_selection import ParameterGrid

class BaseCrossVal(ABC):
    """Base class for crossval: kfold."""
    
    @abstractmethod
    def __init__(self, model, X, Y, param_dict, folds=10, bootnum=100):
        self.model = model 
        self.X = X
        self.Y = Y
        self.param_dict = param_dict
        self.param_list = list(ParameterGrid(param_dict))
        self.folds = folds
        self.bootnum = bootnum
        self.num_param = len(param_dict)
    
    @abstractmethod
    def calc_ypred(self):
        """Calculates ypred full and ypred cv."""
        pass
    
    @abstractmethod
    def calc_stats(self):
        """Calculates binary statistics from ypred full and ypred cv."""
        pass
    
    @abstractmethod
    def run(self):
        """Runs all functions prior to plot."""
        pass
    
    @abstractmethod
    def plot(self):
        """Creates a R2/Q2 plot."""
        pass